package Backpropagation;

import java.util.Vector;

//Represents a Backpropagation network.
//Implements:
//--Train
//--Test 
//--Classification test
//--Forward and Backward passes
//--Learning
//--Getting inputs and evaluating outputs with Errors
//--K-fold cross validation (and remembers the instances)
//----Other
public class Backpropagation {
	private double alpha;
	private double mi;
	
	//Graph
	private Vector<Layer> network; 
	
	//Instance from the k-fold validation - e_val 
	private Vector<Model> network_instances; 
	
	
	public Backpropagation(
			int input_dimension,
			int output_dimension, 
			
			//For example [16,32,16] will make a 3-hidden layer network with 16,32 and 16 neurons on each layer. 
			Vector<Integer> layer_size
		){
		
		//Allocate network
		this.network = new Vector<Layer>();
		this.network_instances = new Vector<Model>();
		//input + hidden + output
		int layer_count = layer_size.size();
		this.network.setSize(layer_count + 2);
		
		
		//Note: + 1 stands for threshold
		//input layer
		this.network.set(0, 
				new Layer(input_dimension + 1));
		//output layer
		this.network.set(this.network.size()-1, 
				new Layer(output_dimension));
		//hidden layers
		for(int i=1; i < this.network.size()-1 ; i++){
			this.network.set(i, new Layer(layer_size.get(i-1)+1));
		}
		
		//setup Elements
		//inputLayer
		for(int j=0; j<this.inputLayer().size() ; j++){
			this.inputLayer().set(j, new Element());
		}
		//Hidden layers (with input weights and treshold)
		for(int i=1; i<this.network.size()-1 ; i++){
			for(int j=0; j<this.network.get(i).size() - 1 ; j++){
			  this.network.get(i).set(j, 
					  new Element(this.network.get(i-1).size())
			  );
			}
			//Last element is the threshold
			this.network.get(i).set(this.network.get(i).size()-1, new Element());
		}
		
		for(int j=0; j<this.outputLayer().size() ; j++){
			this.outputLayer().set(j, new Element(
					this.network.get(this.network.size()-2).size() //Size of the layer before the last
			));
		}
		
		//Setup thresholds (on all except output layer)
		for(int i=0; i< this.network.size() - 1 ; i++){
			this.network.get(i).getLast().y=1.0;
		}
	}
	
	public void setAlpha(double _alpha){
		this.alpha = _alpha;
	}
	public void setMomentum(double _mi){
		this.mi = _mi;
	}
	
	//Set y (outputs) on the input neurons (Elements)
	protected void setUpInputs(Supervisor supervisor){
		//!DEVELOPER
		if(supervisor.getInputDimension() + 1 != this.inputLayer().size())
			System.out.println("Throw incompatible input dimensions");
		
		for(int j=0; j<supervisor.getInputDimension() ; j++){
			this.inputLayer().get(j).y = supervisor.getInputByIndex(j);
		}
	}
	//Forward pass
	protected void setUpOutputs(){
		//For all weighted neurons
		for(int i=1 ; i < this.network.size() ; i++){
			for(int j=0; j<this.network.get(i).size() ; j++){
				//Threshold "output computation" is skipped internally
				this.network.get(i).get(j).computeY(this.network.get(i-1));
			}
		}
	}
	//Backward pass
	protected void setUpDeltas(Supervisor supervisor){
		//!DEVELOPER
		if(supervisor.getOutputDimension() != this.outputLayer().size())
			System.out.println("Throw incompatible output dimensions");
		
		//Output layer gets target outputs.
		for(int j=0; j < this.outputLayer().size() ; j++){
			this.outputLayer().get(j).computeDelta(
				supervisor.getOutputByIndex(j)	
			);
		}
		
		//From last to first HiddenLayer update with Backpropagation (Backward pass) Rule
		for(int i=this.network.size()-2 ; i >= 1 ; i--){
			for(int j=0; j<this.network.get(i).size() ; j++){
				//Thresholds are skipped internally
				this.network.get(i).get(j).computeDelta(
						this.network.get(i+1),
						j
				);
			}
		}
	}
	//Learning procedure (also implements momentum -- mi (coefficient))
	protected void updateWeights(){
		for(int i=1 ; i<this.network.size() ; i++){
			for(int j=0; j<this.network.get(i).size() ; j++){
				//Thresholds are skipped internally
				this.network.get(i).get(j).learn(this.network.get(i-1), this.alpha, this.mi);
			}
		}
	}
	
	//Returns avarage error
	//Tests this network on 
	public double testNetwork(Supervisor tester, boolean log){
		double total_error = 0.0;
		
		do{
			this.setUpInputs(tester);
			//Forward pass / compute outputs. 
			this.setUpOutputs();
			double d_error = this.getError(tester);
			
			if(log) 
				System.out.println(d_error);
			
			total_error += d_error;
		}while(tester.next());
		
		return total_error /= ((Double)(double)tester.getSize());
	}
	
	//Returns classification error in percents
	public double testClassification(Supervisor tester, boolean log_classification, boolean log_confusion_matrix){
		int error_count = 0;
		
		int cm_size = this.getOutputs().size();
		Integer[][] CM = new Integer[cm_size][cm_size]; 
		for(int i=0; i<cm_size;i++)
			for(int j=0; j<cm_size;j++)
				CM[i][j] = 0;
		
		do{
			this.setUpInputs(tester);
			//Forward pass / compute outputs. 
			this.setUpOutputs();
			
			//Wrong classification
			Integer target = tester.classify(tester.getCurrent().output);
			Integer output = tester.classify(this.getOutputs());
			
			CM[target][output]++;
			
			if(log_classification){
				for(Double x : tester.getCurrent().input)
					System.out.print(x.toString() + " ");
				System.out.println(target.toString() + " " + output.toString());
			}
			
			if(target != output){
				error_count++;
			}
		}while(tester.next());
		
		if(log_confusion_matrix){
			for(int i=0; i<cm_size;i++){
				for(int j=0; j<cm_size;j++){
					System.out.print(CM[i][j].toString() + " ");
				}
				System.out.println();
			}
		}
		
		return ((Double)(double)error_count) / ((Double)(double)tester.getSize()) * 100.0;
	}
	
	public Vector<Double> getOutputs(){
		Vector<Double> result = new Vector<Double>();
		for(Element e : this.outputLayer()){
			result.add(e.y);
		}
		return result; 
	}
	
	//Train on one input-output entry. 
	public void trainNetwork(Supervisor trainer){
		this.setUpInputs(trainer);
		this.setUpOutputs();
		this.setUpDeltas(trainer);
		this.updateWeights();
	}
	
	//K-cross validation 
	//--remembers each instance and its average error on validation set. 
	public double trainNetwork(Supervisor supervisor, int epoch_count, int k, boolean log){
		double global_cv = 0.0; 
		
		for(int i=0; i<k; i++){
			Supervisor estimator = supervisor.extractComplement(i, i+1, k);
			Supervisor validator = supervisor.extract(i, i+1, k);
			
			System.out.println("Test case: " + ((Integer)i).toString() + " of k="  + ((Integer)k).toString() );
			double e_val = this.trainNetwork(
					epoch_count,
					estimator,
					validator,
					log
			);
			//Remember this network. 
			this.addNetworkToInstances(estimator, validator, e_val);
			this.resetNetwork();
			global_cv += e_val;
		}
		
		return global_cv / ((double)k);
	}
	/*
	public Double getEstimationError(){
		return this.last_estimation_error;
	}
	public Double getValidationError(){
		return this.last_validation_error;
	}*/
	
	private static double avg(Vector<Double> v, int from, int to){
		double result = 0.0;
		for(int i = from; i < to ; i++){
			result += v.get(i);
		}
		return result / ((double)(to - from));
	}
	
	//returns error when stopped 
	public double trainNetwork(int epoch_count, Supervisor estimator, Supervisor validator, boolean log){
		//estimator.printEntries();
		//System.out.println("==============");
		//validator.printEntries();
		
		Vector<Double> v_errors = new Vector<Double>();
		//errors will be avaraged in every 10 epochs, if increased, then stop
		//--- or epoch count reached 
		int avg_of_last = 20;
		
		double avg_v_error=-1.0;
		double avg_e_error=-1.0;
		
		for(int ei = 1; ei <= epoch_count ; ei++){
			double e_error = 0.0; //Estimator
			  
			do{
			  this.trainNetwork(estimator);
			  //this.printLog();
			  //this.printNetwork();
			    
			  e_error += this.getError(estimator);
			  //System.out.println("  " + ((Double)B.getError()).toString());
			}while(estimator.next());
			  
			estimator.permutateRandom();
			  
			avg_e_error = e_error / ((double)estimator.getSize());
			avg_v_error = this.testNetwork(validator, false);
			v_errors.add(avg_v_error);
			  
			/*
			if(v_error / ((double)validator.getSize()) < 0.01){
				System.out.println("Stopping criteria met");
				break;
			}*/
			//Stopping criteria
			if(ei % avg_of_last == 0 && ei >= 2 * avg_of_last){
				double last_2 = Backpropagation.avg(v_errors, ei-2*avg_of_last, ei-avg_of_last);
				double last_1 = Backpropagation.avg(v_errors, ei-1*avg_of_last, ei);
				
				//System.out.println(((Double)last_2).toString() + " " + ((Double)last_1).toString());
				
				if(last_2 < last_1){
					System.out.println("Stopping criteria met at "+((Integer)ei).toString()+" (avg of last "+((Integer)avg_of_last).toString()+")");
					return avg_v_error;
				}
			}
			if(log){
				System.out.println(
						((Integer)ei).toString() +
						" " +
						((Double)avg_e_error).toString() 
						+ " " 
						+ ((Double)avg_v_error).toString()
				);
			}
		}
		return avg_v_error;
	}
	
	public void resetNetwork(){
		for(int i = 0; i < this.network.size() ; i++){
			for(int j=0; j< this.network.get(i).size() ; j++){
				this.network.get(i).get(j).resetWeights();
			}
		}
	}

	//Also calculates errors on validation networks (with classification errors)
	protected void addNetworkToInstances(Supervisor estimator, Supervisor validator, double e_val){
		
		Vector<Layer> clone = new Vector<Layer>();
		clone.setSize(this.network.size());
		for(int i=0 ; i<this.network.size() ; i++){
			clone.set(i, new Layer(this.network.get(i)));
		}
		
		Model model = new Model();
		model.network = clone;
		model.e_val = e_val;
		model.c_e_est = this.testClassification(estimator, false, false);
		model.c_e_val = this.testClassification(validator, false, false);
		
		model.printLine(true);
		
		this.network_instances.add(model);
	}
	public void setBestInstance(){
		if(this.network_instances.isEmpty())
			return;
		
		Double best_e_val = null;
		for(Model m : this.network_instances){
			if(best_e_val== null || best_e_val > m.e_val){
				best_e_val = m.e_val;
				this.network = m.network;
			}
		}
	}

	protected Vector<Element> inputLayer(){
		return this.network.get(0);
	}
	protected Vector<Element> outputLayer(){
		return this.network.get(this.network.size() - 1);
	}
	

	public double getError(Supervisor supervisor){
		if(supervisor.getOutputDimension() != this.outputLayer().size())
			System.out.println("Throw incompatible output dimensions (in error computation)");
		
		double result = 0.0;
		for(int j=0; j < this.outputLayer().size() ; j++){
			result += Math.pow(supervisor.getOutputByIndex(j) - this.outputLayer().get(j).y, 2);
		}
		return result; 
	}
	
	public void printLog(Supervisor supervisor){
		for(int j=0; j<this.inputLayer().size() ; j++)
			System.out.print(((Double)this.inputLayer().get(j).y).toString() + " ");
		System.out.println(" -- inputs");
		
		for(int j=0; j<this.outputLayer().size() ; j++)
			System.out.print(((Double)this.outputLayer().get(j).y).toString() + " ");
		System.out.println(" -- outputs");
		
		for(int j=0; j<supervisor.getOutputDimension() ; j++){
			System.out.print(((Double)supervisor.getOutputByIndex(j)).toString() + " ");
		}
		System.out.println(" -- targets");
		
		System.out.println(((Double)this.getError(supervisor)).toString() + " -- error");
	}
	
	public void printNetwork(){
		for(int i=0; i < this.network.size() ; i++){
			System.out.println("Layer id=" + ((Integer)i).toString());
			
			Layer layer = this.network.get(i);
			for(int j=0; j < layer.size() ; j++){
				System.out.println("y=" + ((Double)(layer.get(j).y)).toString() + " delta=" + ((Double)(layer.get(j).delta)).toString());
				
				Vector<Double> weight = layer.get(j).weight;
				if(weight == null)
					continue;
				
				for(int k=0; k<weight.size() ; k++){
					System.out.print(weight.get(k).toString() + " ");
				}
				System.out.println();
			}
		}
	}
	
}
